#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

__docformat__ = 'restructuredtext'


# import standard library
from datetime import datetime

# import interfaces
from pyams_security.interfaces import ISecurityManager
from pyams_utils.interfaces import VIEW_PERMISSION
from pyams_workflow.interfaces import IWorkflowManagedContent, IWorkflowPublicationInfo, IWorkflow, IWorkflowVersions, \
    IWorkflowPublicationSupport, IObjectClonedEvent, IWorkflowState, VersionError, VERSION_DISPLAY, \
    DISPLAY_FIRST_VERSION, DISPLAY_CURRENT_VERSION
from zope.dublincore.interfaces import IZopeDublinCore

# import packages
from persistent import Persistent
from pyams_utils.adapter import adapter_config, get_annotation_adapter
from pyams_utils.date import format_datetime, format_date, SH_DATE_FORMAT
from pyams_utils.registry import get_utility
from pyams_utils.request import check_request
from pyams_utils.timezone import gmtime, tztime, GMT
from pyams_utils.traversing import get_parent
from pyams_utils.vocabulary import vocabulary_config
from pyams_workflow.versions import WorkflowHistoryItem
from pyramid.events import subscriber
from zope.container.contained import Contained
from zope.interface import implementer
from zope.schema.fieldproperty import FieldProperty
from zope.schema.vocabulary import SimpleVocabulary, SimpleTerm

from pyams_workflow import _


@vocabulary_config(name='PyAMS content publication date')
class WorkflowContentDisplayedDateVocabulary(SimpleVocabulary):
    """Workflow content displayed date vocabulary"""

    def __init__(self, context):
        request = check_request()
        terms = []
        versions = IWorkflowVersions(context.__parent__)
        # check for first version
        first_version_label = request.localizer.translate(VERSION_DISPLAY[DISPLAY_FIRST_VERSION])
        try:
            first_version = versions.get_version(1)
        except VersionError:
            pass
        else:
            info = IWorkflowPublicationInfo(first_version, None)
            if info is not None and info.publication_effective_date:
                first_version_label = '{1} (= {0})'.format(first_version_label.lower(),
                                                           format_date(info.publication_effective_date,
                                                                       format=SH_DATE_FORMAT))
        terms.append(SimpleTerm(DISPLAY_FIRST_VERSION, title=first_version_label))
        # check for current version
        current_version_label = request.localizer.translate(VERSION_DISPLAY[DISPLAY_CURRENT_VERSION])
        info = IWorkflowPublicationInfo(context, None)
        if info is not None and info.publication_effective_date:
            current_version_label = '{1} (= {0})'.format(current_version_label.lower(),
                                                         format_date(info.publication_effective_date,
                                                                     format=SH_DATE_FORMAT))
        terms.append(SimpleTerm(DISPLAY_CURRENT_VERSION, title=current_version_label))
        super(WorkflowContentDisplayedDateVocabulary, self).__init__(terms)


@implementer(IWorkflowPublicationInfo)
class WorkflowContentPublicationInfo(Persistent, Contained):
    """Workflow content info"""

    _publication_date = FieldProperty(IWorkflowPublicationInfo['publication_date'])
    publisher = FieldProperty(IWorkflowPublicationInfo['publisher'])
    _first_publication_date = FieldProperty(IWorkflowPublicationInfo['first_publication_date'])
    _publication_effective_date = FieldProperty(IWorkflowPublicationInfo['publication_effective_date'])
    push_end_date = FieldProperty(IWorkflowPublicationInfo['push_end_date'])
    _publication_expiration_date = FieldProperty(IWorkflowPublicationInfo['publication_expiration_date'])
    displayed_publication_date = FieldProperty(IWorkflowPublicationInfo['displayed_publication_date'])

    @property
    def publication_date(self):
        return self._publication_date

    @publication_date.setter
    def publication_date(self, value):
        self._publication_date = gmtime(value)

    @property
    def publication(self):
        request = check_request()
        translate = request.localizer.translate
        auth = get_utility(ISecurityManager)
        return translate(_('{date} by {principal}')).format(
            date=format_datetime(tztime(IWorkflowPublicationInfo(self).publication_date), request=request),
            principal=translate(auth.get_principal(self.publisher).title))

    @property
    def first_publication_date(self):
        return self._first_publication_date

    @property
    def publication_effective_date(self):
        return self._publication_effective_date

    @publication_effective_date.setter
    def publication_effective_date(self, value):
        value = gmtime(value)
        if value:
            if (self._first_publication_date is None) or (self._first_publication_date > value):
                self._first_publication_date = value
            IZopeDublinCore(self.__parent__).effective = value
        elif self._publication_effective_date:
            del IZopeDublinCore(self.__parent__)._mapping['Date.Effective']
        self._publication_effective_date = value

    @property
    def push_end_date_index(self):
        value = self.push_end_date or self.publication_expiration_date
        if value is None:
            value = datetime(9999, 12, 31, 11, 59, 59, 999999, GMT)
        return gmtime(value)

    @property
    def publication_expiration_date(self):
        return self._publication_expiration_date

    @publication_expiration_date.setter
    def publication_expiration_date(self, value):
        value = gmtime(value)
        if value:
            IZopeDublinCore(self.__parent__).expires = value
        elif self._publication_expiration_date:
            del IZopeDublinCore(self.__parent__)._mapping['Date.Expires']
        self._publication_expiration_date = value

    def reset(self, complete=True):
        self._publication_date = None
        self._first_publication_date = None
        self._publication_effective_date = None
        if complete:
            self._publication_expiration_date = None
            self.push_end_date = None

    def is_published(self, check_parent=True):
        # check is parent is published
        if check_parent:
            parent = get_parent(self.__parent__, IWorkflowPublicationSupport, allow_context=False)
            if (parent is not None) and not IWorkflowPublicationInfo(parent).is_published():
                return False
        # associated workflow?
        workflow = IWorkflow(self.__parent__, None)
        if (workflow is not None) and not workflow.published_states:
            return False
        # check content versions
        versions = IWorkflowVersions(self.__parent__, None)
        if (versions is not None) and not versions.get_versions(workflow.published_states):
            return False
        now = tztime(datetime.utcnow())
        return (self.publication_effective_date is not None) and \
               (self.publication_effective_date <= now) and \
               ((self.publication_expiration_date is None) or
                (self.publication_expiration_date >= now))

    def is_visible(self, request=None, check_parent=True):
        # associated workflow?
        content = IWorkflowManagedContent(self.__parent__, None)
        if content is not None:
            wf_name = content.workflow_name
            if not wf_name:
                return True
            # check workflow?
            if content.view_permission:
                if request is None:
                    request = check_request()
                if not request.has_permission(content.view_permission, context=self.__parent__):
                    return False
        else:
            if not request.has_permission(VIEW_PERMISSION, context=self.__parent__):
                return False
        return self.is_published(check_parent)


WORKFLOW_CONTENT_KEY = 'pyams_workflow.content_info'


@adapter_config(context=IWorkflowPublicationSupport, provides=IWorkflowPublicationInfo)
def workflow_content_publication_info_factory(context):
    """Workflow content info factory"""
    return get_annotation_adapter(context, WORKFLOW_CONTENT_KEY, WorkflowContentPublicationInfo)


@subscriber(IObjectClonedEvent)
def handle_cloned_object(event):
    """Add comment when an object is cloned"""
    request = check_request()
    translate = request.localizer.translate
    source_state = IWorkflowState(event.source)
    item = WorkflowHistoryItem(date=datetime.utcnow(),
                               principal=request.principal.id,
                               comment=translate(_("Clone created from version {source} (in « {state} » state)")).format(
                                   source=source_state.version_id,
                                   state=translate(IWorkflow(event.source).get_state_label(source_state.state))))
    target_state = IWorkflowState(event.object)
    target_state.history.clear()
    target_state.history.append(item)


@subscriber(IObjectClonedEvent, context_selector=IWorkflowPublicationSupport)
def handle_cloned_publication_support(event):
    """Reset publication info when an object is cloned"""
    IWorkflowPublicationInfo(event.object).reset(complete=False)
